<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: 9548f1f81a43 $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2020-07-02 12:51 +0200 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */
class we_shop_vatRule{
	/** @var string */
	public $stateField;

	/** @var string */
	var $stateFieldIsISO;

	/** @var string[] */
	public $liableToVat = [];

	/** @var string[] */
	public $notLiableToVat = [];

	/** @var array<int,array{states:array,customerField:string,condition:string,returnValue:bool}>
	 * a rule is an array containing
	 */
	//		states -> array
	//		customerField -> string
	//		condition -> string
	//		returnValue -> 0 | 1
	public $conditionalRules;

	/** @var string */
	public $defaultValue = 'true';

	/** @var array<int,array{states:array,customerField:string,condition:string,returnValue:bool}> */
	private static $emptyConditional = [['states' => [],
		'customerField' => '',
		'condition' => '',
		'returnValue' => true
	]];

	private function __construct(string $defaultValue, string $stateField, array $liableToVat, array $notLiableToVat, array $conditionalRules, string $stateFieldIsISO = '0'){
		$this->defaultValue = $defaultValue;
		$this->stateField = $stateField;
		$this->stateFieldIsISO = $stateFieldIsISO;
		$this->liableToVat = $liableToVat;
		$this->notLiableToVat = $notLiableToVat;
		$this->conditionalRules = $conditionalRules;
	}

	public function executeVatRule(array $customer = [], string $state = ''): bool{
		// now check all rules for the vat

		if($customer || $state){
			if((isset($customer[$this->stateField]) || $state)){
				$state = (string) ($customer[$this->stateField] ?? ($state ?: ''));

				// is state liableToVat
				if(in_array($state, $this->liableToVat)){
					return true;
				}

				// is state not liable to vat
				if(in_array($state, $this->notLiableToVat)){
					return false;
				}

				// if we have $state but no $customer: return here
				if(!$customer){
					return ($this->defaultValue === 'true');
				}

				// now check additional fields
				foreach($this->conditionalRules as $rule){
					$ret = ($rule['returnValue'] === 'true');
					$field = $rule['customerField'] ? $customer[$rule['customerField']] : '';

					if(in_array($state, $rule['states'])){
						switch($rule['condition']){
							case 'is_empty':
								return (empty($field) ? $ret : !$ret);
							case 'is_set':
								return (!empty($field) ? $ret : !$ret);
						}
					}
				}
			}
		}

		return ($this->defaultValue === 'true');
	}

	public static function initByRequest(): self{
		return new self(
			we_base_request::string('defaultValue'),
													 we_base_request::string('stateField'),
																			self::resolveLineBreaksInArray(we_base_request::stringList('liableToVat')),
																																							self::resolveLineBreaksInArray(we_base_request::stringList('notLiableToVat')),
																																																	self::makeArrayFromConditionField(),
																																																	we_base_request::string('stateFieldIsISO')
		);
	}

	public static function getShopVatRule(): self{
		if(($strfelder = fS('SELECT pref_value FROM ' . SETTINGS_TABLE . ' WHERE tool="shop" AND pref_name="weShopVatRule"'))){
			$felder = we_unserialize($strfelder);
			return new self($felder['defaultValue'],
									 $felder['stateField'],
									 $felder['liableToVat'],
									 $felder['notLiableToVat'],
									 $felder['conditionalRules'],
									 $felder['stateFieldIsISO']
			);
		}
		return new self('true', '', [], [], self::$emptyConditional, '');
	}

	public static function getStateField(): string{
		$rule = self::getShopVatRule();
		return $rule->stateField;
	}

	private static function makeArrayFromConditionField(): array{
		/*
		 * $conditionalRules once was intended to be an array of different rules for different sets of countries,
		 * but this was never fully implemented in vatRules dialog: this method is now ready to process multiple rules
		 */

		$ret = [];
		$customerFields = we_base_request::stringList('conditionalCustomerField');
		$conditions = we_base_request::stringList('conditionalCondition');
		$returnValues = we_base_request::stringList('conditionalReturn');

		foreach(we_base_request::stringList('conditionalStates') as $i => $states){
			if($states){
				$ret[] = [
					'states' => self::resolveLineBreaksInArray(we_base_request::filterStringList($states)),
					'customerField' => $customerFields[$i],
					'condition' => $conditions[$i],
					'returnValue' => $returnValues[$i],
				];
			}
		}

		return $ret ?: self::$emptyConditional;
	}

	/**
	 *
	 * @param string[] $arr
	 * @return string[]
	 */
	private static function resolveLineBreaksInArray(array $arr): array{
		$ret = [];
		foreach($arr as $lines){
			$ret = array_merge($ret, array_unique(array_filter(array_map('trim', explode("\n", $lines)))));
		}

		return $ret;
	}

	function save(): bool{
		$DB_WE = DB_WE::$inst;

		if($DB_WE->query('REPLACE ' . SETTINGS_TABLE . ' set pref_value="' . $DB_WE->escape(we_serialize((array) $this, SERIALIZE_JSON)) . '",tool="shop",pref_name="weShopVatRule"')){
			$CLFields = we_unserialize(fS('SELECT pref_value FROM ' . SETTINGS_TABLE . ' WHERE tool="shop" AND pref_name="shop_CountryLanguage"'));
			if(!$CLFields){
				$CLFields = ['stateField' => $this->stateField,
					'stateFieldIsISO' => $this->stateFieldIsISO
				];
				$DB_WE->query('UPDATE ' . SETTINGS_TABLE . ' SET pref_value="' . $DB_WE->escape(we_serialize($CLFields, SERIALIZE_JSON)) . '" WHERE tool="shop",pref_name="shop_CountryLanguage"');
			}

			return true;
		}
		return false;
	}

}
