<?php
declare(strict_types=1);

/* webEdition CMS
 *
 * $Rev: 9b90455e1eb9 $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2020-11-23 12:06 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */

abstract class we_weTag_tag{
	public const DIR = 'tag';

	/** @var we_document_webEdition */
	protected static $doc = null;

	/** @var string added for blockElements */
	protected static $postTagName = '';

	/**
	 * @param array<string,mixed> $attribs parsed tag-attributes
	 * @param string $content content between openingtag and closing tag
	 * @param bool $internal if used by we internally
	 * @throws we_exception_parser if any attribute is missing
	 */
	abstract public static function tag(array $attribs = [], string $content = '', bool $internal = false);

	/**
	 * @param array<string,mixed> $attribs parsed tag-attributes
	 * @param string $content content between openingtag and closing tag
	 * @return string parsed tag for php
	 */
	public static function parse(array $attribs, string $content): string{
		$tagname = strtr(get_called_class(), ['we_tag_' => '']);
		return '<?= ' .
			($content ?
			// Tag besitzt Endtag
			we_weTag_parser::printTag($tagname, $attribs, $content, true) . '; ?>' :
			// Tag ohne Endtag
			we_weTag_parser::printTag($tagname, $attribs) . '; ?>'
			);
	}

//tag utility functions

	/**
	 * get an attribute from $attribs, and return its value according to default
	 * @param string $name attributes name
	 * @param array<string,mixed> $attribs array containg the attributes
	 * @param mixed $default default value
	 * @return mixed returns the attributes value or default if not set
	 */
	public static function getParserAttribute(string $name, array $attribs, /* mixed */ $default, string $type, bool $isDeprecated = false)/* :mixed */{
		return self::getAttribute($name, $attribs, $default, $type, $isDeprecated, false);
	}

	/**
	 * get an attribute from $attribs, and return its value according to default
	 * @param string $name attributes name
	 * @param array<string,mixed> $attribs array containg the attributes
	 * @param mixed $default default value
	 * @param bool $useGlobal check if attribute value is a php-variable and is found in $GLOBALS
	 * @return mixed returns the attributes value or default if not set
	 */
	public static function getAttribute(string $name, array $attribs, /* mixed */ $default, string $type, bool $isDeprecated = false, bool $useGlobal = true)/* :mixed */{
		if(!isset($attribs[$name])){
			return is_string($default) ? htmlspecialchars_decode($default) : $default;
		}
		$regs = [];
		if($isDeprecated){
			t_e('deprecated', 'Attribute ' . $name . ' is deprecated and will be removed. Please update your code!');
		}
		if($useGlobal && is_string($attribs[$name]) && preg_match('|^\\\\?\$([^\[]+)(\[.*\])?|', $attribs[$name], $regs)){
			$value = (isset($regs[2]) ?
				self::getArrayValue($GLOBALS, $regs[1], $regs[2]) :
				($GLOBALS[$regs[1]] ?? ''));
		} else {
			$value = $attribs[$name];
		}

		$value = we_base_request::filterVar($value, $type, $default); //@phan-suppress-current-line PhanDeprecatedFunction

		return is_string($value) ? htmlspecialchars_decode($value) : $value;
	}

	/**
	 * get an attribute from $attribs, and return its value according to default; unset attribs after access
	 * @param string $name attributes name
	 * @param array<string,mixed> $attribs array containg the attributes
	 * @param mixed $default default value
	 * @param bool $useGlobal check if attribute value is a php-variable and is found in $GLOBALS
	 * @return mixed returns the attributes value or default if not set
	 */
	protected static function getAttributeR(string $name, array &$attribs, $default = '', string $type = we_base_request::RAW, bool $isDeprecated = false, bool $useGlobal = true)/* :mixed */{
		if(!isset($attribs[$name])){
			return $default;
		}
		$ret = self::getAttribute($name, $attribs, $default, $type, $isDeprecated, $useGlobal);
		unset($attribs[$name]);
		return $ret;
	}

	public static function parseError(string $text, string $extra = ''): string{
		if(defined('IGNORE_PARSE_ERROR')){
			return '';
		}
		t_e('warning', html_entity_decode($text, ENT_QUOTES, 'UTF-8'), g_l('weClass', '[template]') . ': ' . we_weTag_parser::$curFile, $extra);
		return '<b>' . g_l('parser', '[error_in_template]') . ':</b>' . $text . "<br/>\n" . g_l('weClass', '[template]') . ': ' . we_weTag_parser::$curFile;
	}

	private static function handleTagError(string $text, string $file): string{
		if(defined('IGNORE_PARSE_ERROR')){
			return '';
		}
		t_e('warning', html_entity_decode($text, ENT_QUOTES, 'UTF-8'), g_l('weClass', '[template]') . ': ' . $file);
		return '<b>' . g_l('parser', '[error_in_template]') . ':</b>' . $text . "<br/>\n" . g_l('weClass', '[template]') . ': ' . $file;
	}

	/**
	 * @param array<string,mixed> $attribs
	 * @param array<string,bool> $attrs
	 * @param string $tag
	 * @throws we_exception_parser
	 */
	public static function attributeMissing(array $attribs, array $attrs, string $tag): void{//FIXME: this should throw an error without return value
		$tag = str_replace(['we_tag_', 'we_parse_tag_'], '', $tag);

		foreach($attrs as $attr => $canBeEmpty){
			if($canBeEmpty){
				if(!isset($attribs[$attr])){
					throw new we_exception_parser(self::handleTagError(sprintf(g_l('parser', '[attrib_missing2]'), $attr, $tag), we_weTag_parser::$curFile));
				}
			} elseif(!isset($attribs[$attr]) || $attribs[$attr] === ''){
				throw new we_exception_parser(self::handleTagError(sprintf(g_l('parser', '[attrib_missing]'), $attr, $tag), we_weTag_parser::$curFile));
			}
		}
	}

	public static function redirectTagOutput(string $returnvalue, string $nameTo, string $to = 'screen'): string{
		switch(we_tag_calculate::$active ? 'calculate' : $to){
			case 'request':
				self::setVarArray($_REQUEST, $nameTo, $returnvalue);
				return '';
			case 'post':
				self::setVarArray($_POST, $nameTo, $returnvalue);
				return '';
			case 'get':
				self::setVarArray($_GET, $nameTo, $returnvalue);
				return '';
			case 'global':
				self::setVarArray($GLOBALS, $nameTo, $returnvalue);
				return '';
			case 'session':
				self::setVarArray($_SESSION, $nameTo, $returnvalue);
				return '';
			case 'top':
				we_base_WE::$mainDoc['doc']->setElementS($nameTo, $returnvalue);
				return '';
			case 'block' :
				$nameTo = self::getPostName($nameTo);
			case 'self' :
				we_base_WE::getWeDoc()->setElementS($nameTo, $returnvalue);
				return '';
			case 'sessionfield' :
				if(isset($_SESSION['webuser'][$nameTo])){
					$_SESSION['webuser'][$nameTo] = $returnvalue;
				}
				return '';
			case 'calculate':
				return (string) we_base_util::std_numberformat($returnvalue);
			case 'screen':
			default:
				return $returnvalue;
		}
	}

	/**
	 * get the full name of an Attribute with applied postTagName if set
	 * @param string $var
	 * @return string
	 */
	public static function getPostName(string $var): string{
		return $var . self::$postTagName;
	}

	public static function getDocForTag(string $docAttr, bool $maindefault = false): we_document_base{
		switch($docAttr){
			case 'self':
				return we_base_WE::getWeDoc();
			case 'top':
				$doc = we_base_WE::$mainDoc['doc'];
				assert($doc !== null);
				return $doc;
			default:
				$doc = we_base_WE::$mainDoc['doc'];
				assert($doc !== null);
				return ($maindefault ?
					$doc :
					we_base_WE::getWeDoc());
		}
	}

	private static function setVarArray(array &$arr, string $string, $value): void{
		if(!str_contains($string, '[')){
			$arr[$string] = $value;
			return;
		}
		$current = &$arr;

		$arr_matches = preg_split('/\]\[|\[|\]/', $string);
		$last = count($arr_matches) - 1;
		unset($arr_matches[$last--]); //preg split has an empty element at the end
		foreach($arr_matches as $pos => $dimension){
			if(empty($dimension)){
				$dimension = count($current);
			}
			if($pos == $last){
				$current[$dimension] = $value;
				return;
			}
			if(!isset($current[$dimension])){
				$current[$dimension] = [];
			}
			$current = &$current[$dimension];
		}
	}

	/**
	 *
	 * @param array<string,mixed> $attribs
	 * @param bool $fallBackBackend
	 * @return array{lang:string,code:string}
	 */
	public static function getOutLang(array $attribs, bool $fallBackBackend = false): array{
		$lang = self::getAttribute('outputlanguage', $attribs, '', we_base_request::STRING);
		if(!$lang){
			$doc = self::getDocForTag(self::getAttribute('doc', $attribs, 'self', we_base_request::STRING));
			$lang = $doc->Language;
		}
		if(!$lang){
			$lang = (string) ($fallBackBackend ? array_search(we_base_WE::$Language, we_base_country::getWELangs(), true) : we_base_WE::$Langs['DefaultFrontendLanguage']) ?: 'de';
		}
		$langcode = substr($lang, 0, 2);
		return ['lang' => $lang, 'code' => $langcode];
	}

	public static function modulMissingError(string $modul, string $tag): string{
		return self::handleTagError(sprintf(g_l('parser', '[module_missing]'), $modul, str_replace(['we_tag_', 'we_parse_tag_'], '', $tag)), fS('SELECT Path FROM ' . TEMPLATES_TABLE . ' WHERE ID=' . we_base_WE::getWeDoc()->TemplateID));
	}

	protected static function deprecated(string $since){
		$tagname = strtr(get_called_class(), ['we_tag_' => '']);
		t_e('call of deprecated tag we:"' . $tagname . '" which is deprecated since ' . $since . ' will be removed soon.');
	}

	public static function getArrayValue($var, $name, string $arrayIndex, bool $isset = false)/* :mixed */{
		$arr_matches = preg_split('/\]\[|\[|\]/', $arrayIndex);
		if(count($arr_matches) > 1){
			unset($arr_matches[count($arr_matches) - 1]);
		}
		if($name !== null){
			$arr_matches[0] = $name;
		}
		foreach($arr_matches as $cur){
			if(!isset($var[$cur])){
				return ($isset ? false : '');
			}
			$var = $var[$cur];
		}
		return ($isset ? true : $var);
	}

	private static function cutSimpleText(string $text, int $len): string{
		if($len >= strlen($text)){
			return $text;
		}
		$text = substr($text, 0, $len);
		//cut to last whitespace, if any.
		return substr($text, 0, max([
				strrpos($text, ' '),
				strrpos($text, '.'),
				strrpos($text, ','),
				strrpos($text, '/'),
				strrpos($text, "\n"),
				strrpos($text, "\t"),
			]) ?: $len
		);
	}

	public static function cutText(string $text, int $max = 0, bool $striphtml = false): string{
		$text = $striphtml ? strip_tags($text) : $text;
		if((!$max) || (strlen($text) <= $max)){
			return $text;
		}
		//no tags, simple cut off
		if($striphtml || !str_contains($text, '<')){
			return self::cutSimpleText($text, $max) . ($striphtml ? ' ...' : ' &hellip;');
		}

		$ret = '';
		$tags = $foo = [];
		//split text on tags, entities and "rest"
		preg_match_all('%(&#?[[:alnum:]]+;)|([^<&]*)|<(/?)([[:alnum:]]+)([ \t\r\n]+[[:alnum:]]+[ \t\r\n]*=[ \t\r\n]*"[^"]*")*[ \t\r\n]*(/?)>%sm', $text, $foo, PREG_SET_ORDER);

		foreach($foo as $cur){
			switch(count($cur)){
				case 2://entity
					if($max > 0){
						$ret .= $cur[0];
						$max -= 1;
					}
					break;
				case 3://text
					if($max > 0){
						$len = strlen($cur[0]);
						$ret .= ($len > $max ? self::cutSimpleText($cur[0], $max) : $cur[0]);
						$max -= $len;
						if($max <= 0){
							$ret .= ' &hellip;';
						}
					}
					break;
				case 7://tags
					if($max > 0){
						$ret .= $cur[0];
						if(!$cur[6]){//!selfclosing
							if($cur[3]){//close
								array_pop($tags);
							} else {
								$tags[] = $cur[4];
							}
						}
					}
					break;
			}
		}

//close open tags
		while($tags){
			$ret .= '</' . array_pop($tags) . '>';
		}

		return $ret;
	}

	public static function inputRadioField(string $name, string $value, string $itsValue, array $atts): string{
		//  This function replaced fnc: we_getRadioField
		$atts['type'] = 'radio';
		$atts['name'] = $name;
		$atts['value'] = htmlspecialchars($itsValue, -1, 'ISO-8859-1', false);
		if($value == $itsValue){
			$atts['checked'] = 'checked';
		}
		return self::htmlTag('input', $atts);
	}

	public static function textareaField(string $name, string $value, array $atts): string{
		$atts['name'] = $name;
		return self::htmlTag('textarea', $atts, htmlspecialchars($value), true);
	}

	public static function inputTextInputField(string $name, string $value, array $atts): string{
		$atts['type'] = 'text';
		$atts['name'] = $name;
		$atts['value'] = htmlspecialchars($value);
		return self::htmlTag('input', $atts);
	}

	public static function inputCheckboxField(string $name, string $value, array $attr): string{
		//  returns a checkbox with associated hidden-field
		$tmpname = md5(uniqid(__FUNCTION__, true)); // #6590, changed from: uniqid(time())
		if($value){
			$attr['checked'] = 'checked';
		}
		$attr['type'] = 'checkbox';
		$attr['value'] = 1;
		$attr['name'] = $tmpname;
		$attr['onclick'] = 'this.form.elements[\'' . $name . '\'].value=(this.checked) ? 1 : 0';

		return self::htmlTag('input', $attr) . self::htmlTag('input', ['type' => 'hidden', 'name' => $name, 'value' => htmlspecialchars($value)]);
	}

	/**
	 *
	 * @param string $name
	 * @param string $value
	 * @param string[] $values
	 * @param array<string,string> $attribs
	 * @param bool $addMissing
	 * @return string
	 */
	public static function selectField(string $name, string $value, array $values, array $attribs = [], bool $addMissing = true): string{
		$attribs['name'] = $name;
		$content = '';
		$isin = 0;
		foreach($values as $option){
			'@phan-var string $option';
			$opt = htmlspecialchars($option, -1, 'ISO-8859-1', false);
			if($option == $value){
				$content .= self::htmlTag('option', ['value' => $option, 'selected' => 'selected'], $opt, true);
				$isin = 1;
			} else {
				$content .= self::htmlTag('option', ['value' => $option], $opt, true);
			}
		}
		if((!$isin) && $addMissing && $value != ''){
			$content .= self::htmlTag('option', ['value' => $value, 'selected' => 'selected'
					], htmlspecialchars($value), true);
		}
		return self::htmlTag('select', $attribs, $content, true);
	}

	/** This function should be used ONLY in generating code for the FRONTEND
	 * @return string the html element
	 * @param	string $element the elements name
	 * @param	array $attribs array of attribs
	 * @param	string $content content of the tag, if applicable
	 * @param	boolean $forceEndTag if it should be closed, even there is no content
	 * @param boolean $onlyStartTag print
	 * @desc	returns the html element with the given attribs.attr[pass_*] is replaced by "*" to loop some
	 *          attribs through the tagParser.
	 */
	public static function htmlTag(string $element, array $attribs = [], string $content = '', bool $forceEndTag = false, bool $onlyStartTag = false): string{
//	remove attributes
		$removeAttribs = ['xml', 'xmltype', 'to', 'nameto', '_name_orig', 'seeMode'];

		switch($element){
			case 'img':
			case 'a':
				if(defined('HIDENAMEATTRIBINWEIMG_DEFAULT') && HIDENAMEATTRIBINWEIMG_DEFAULT && we_base_WE::$mainDoc['doc'] && !we_base_WE::$mainDoc['doc']->InWebEdition){
					$removeAttribs[] = 'name';
				}
				break;
			case 'form':
				if(defined('HIDENAMEATTRIBINWEFORM_DEFAULT') && HIDENAMEATTRIBINWEFORM_DEFAULT && we_base_WE::$mainDoc['doc'] && !we_base_WE::$mainDoc['doc']->InWebEdition){
					$removeAttribs[] = 'name';
				}
				break;
		}

		return '<' . $element . we_html_tools::makeHTMLTagAtts(self::removeAttribs($attribs, $removeAttribs)) . ($content || $forceEndTag ? //	use endtag
			'>' . $content . '</' . $element . '>' :
//	xml style or not
			( (!$onlyStartTag) ? ' />' : '>'));
	}

	/**
	 * @return array
	 * @param array<string,mixed> $attribs
	 * @param string[] $remove
	 * @desc removes all entries of $attribs, where the key from attribs is in values of $remove
	 */
	public static function removeAttribs(array $attribs, array $remove = []): array{
		return array_filter($attribs, function($k) use($remove){
			return !in_array($k, $remove, true);
		}, ARRAY_FILTER_USE_KEY);
	}

}
